#!/bin/sh

# !!!! START OF TUNABLE PARAMETERS SECTION

retention_period=30

# !!! END OF TUNABLE PARAMETERS SECTION


event_log_host=$1
event_log_dir=$2
database_name=$3
database_dir=$4
logopt_stop=$5
mode=$6
archive_dir=$database_dir/archive
bin_path=$(dirname $0)
logopt_path=$bin_path/logopt

if test -z "$mode"
then 
  "$0" "$@" redirected > "$database_dir/${database_name}.on_new_log.log" 2>&1
  return
fi

set -x
echo "$@"

# optimize log
"$logopt_path" -make "$1" "$2" "$3" "$4" "$5" || error_source="Failed to optimize log"

# clean temporary files possibly left by previous run
"$logopt_path" -clean "$database_name" "$database_dir"
"$logopt_path" -clean "$database_name" "$archive_dir"

# find out latest optimized log and corresponding log file
output="$("$logopt_path" -query "$database_name" "$database_dir")" || error_source="Failed to find out latest optimized log"

while IFS=: read prefix name
do
if test "$prefix" = "latest optimized log"; then latest_optimized_log=$name; fi
if test "$prefix" = "continue with log"; then continue_with_log=$name; fi
done << end_logopt_query
$output
end_logopt_query

echo $latest_optimized_log
echo $continue_with_log

# remove unnecessary optimized log files in the database folder
(cd "$database_dir" && find . -type f -name "$database_name.????????????.optlog" -print -o -path . -o -prune|
  while read d
  do 
    if test $(expr $(basename $d) \< $latest_optimized_log) != 0
    then
      rm "$d" || exit 1
    fi
  done) || error_source="Failed to remove log from database folder"

# move unnecessary log files to archive folder
mkdir "$archive_dir"
(cd "$database_dir" && find . -type f -name "$database_name.????????????.log" -print -o -path . -o -prune|
  while read d
  do 
    if test $(expr $(basename $d) \< $continue_with_log) != 0
    then
      mv "$d" "archive/$(basename $d)" || exit 1
    fi
  done) || error_source="Failed to move log to archive folder"


#rem create logopt file based on outdated log files to ensure safe removal of log files
(cd "$archive_dir" && find . -type f -mtime +$retention_period -name "$database_name.????????????.log" -print -o -path . -o -prune)|
  sort -r |
  while read d
  do 
    "$logopt_path" -make-advanced "$1" "$2" "$3" "$4" "$archive_dir" "$(basename $d)" || set error_source=Failed to optimize log
  done || error_source="Failed to optimize log"

#rem find out latest optimized log and corresponding log file
latest_optimized_log=
continue_with_log=
output="$("$logopt_path" -query "$database_name" "$database_dir" "$archive_dir")" || error_source="Failed to find out latest optimized log"

while IFS=: read prefix name
do
if test "$prefix" = "latest optimized log"; then latest_optimized_log=$name; fi
if test "$prefix" = "continue with log"; then continue_with_log=$name; fi
done << end_logopt_query
$output
end_logopt_query

echo $latest_optimized_log
echo $continue_with_log

# remove unnecessary optimized log files from archive folder
(cd "$archive_dir" && find . -type f -name "$database_name.????????????.optlog" -print -o -path . -o -prune|
  while read d
  do 
    if test $(expr $(basename $d) \< $latest_optimized_log) != 0
    then
      rm "$d" || exit 1
    fi
  done) || error_source="Failed to remove log from archive folder"

# remove unnecessary log files from archive folder
(cd "$archive_dir" && find . -type f -name "$database_name.????????????.log" -print -o -path . -o -prune|
  while read d
  do 
    if test $(expr $(basename $d) \< $continue_with_log) != 0
    then
      rm "$d" || exit 1
    fi
  done) || error_source="Failed to remove log from archive folder"

if test -n "$error_source"
then
  "$bin_path/sccorelogcmd" "$event_log_host" "$event_log_dir" WARNING LogRetention 0 "$error_source. Please inspect $database_name.on_new_log.log for more details."
fi
